
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <map>
#include <string>
#include <vector>
#include "getline.h"
#include "AlphaBeta.h"
#include "PenguinSoccer.h"

void do_slides(const Penguin &p, Direction d, std::vector<Move> &v)
{
    assert(d != NO_DIRECTION);
    for (Direction fd = SOUTH; fd <= NORTH; ++fd) {
        if (fd == NO_DIRECTION) continue;
        v.push_back(Move(SLIDE, p.size, d, fd));
    }
}

void get_pseudo_moves(const Board &b, std::vector<Move> &v)
{
    if (b.game_over())
      return;
    const int who = b.attacker;
    for (int s=0; s < 3; ++s) {
        const Penguin &p = b.penguin[who][s];
        if (!p.on_board) {
            /* All you can do is slide onto the board. */
            do_slides(p, who ? SOUTH : NORTH, v);
            do_slides(p, who ? SOUTHWEST : NORTHWEST, v);
            do_slides(p, who ? SOUTHEAST : NORTHEAST, v);
        } else if (!p.standing) {
            /* All you can do is slide in your current direction */
            do_slides(p, p.dir, v);
            /* or stand up. */
            v.push_back(Move(STAND, p.size));
        } else if (b.has_ball(p)) {
            /* If you have the ball, all you can do is kick it. */
            assert(p.standing);
            for (Direction d = SOUTH; d <= NORTH; ++d) {
                if (d == NO_DIRECTION) continue;
                v.push_back(Move(KICK, p.size, d));
            }
        } else {
            /* If you are standing, you can slide.
             * But let's assume that you only care about
             * sliding in the direction of the ball. */
            Pos bpos = (b.ball_on_board ? b.ball_pos : Pos(3,3));
            for (Direction d = SOUTH; d <= NORTH; ++d) {
                if (d == NO_DIRECTION) continue;
                Pos ps = p.pos;
                ps += d;
                if (ps.manhattan_distance(bpos) > p.pos.manhattan_distance(bpos))
                  continue;
                do_slides(p, d, v);
            }
        }
    }
}

void get_all_moves(const Board &b, std::vector<Move> &v)
{
    assert(v.empty());
    if (b.game_over()) return;
    std::vector<Move> psv;
    get_pseudo_moves(b, psv);
    std::map<std::string, Move> m;
    for (int i=0; i < (int)psv.size(); ++i) {
        Board b2 = b;
        if (apply_move(b2, psv[i])) {
            if (b2.game_over() && b2.ball_pos.isGoal(b.attacker)) {
                /* If this is a winning move, then it's all we need. */
                v.push_back(psv[i]);
                return;
            }
            std::string key = b2.toComparableString();
            std::map<std::string, Move>::iterator it = m.find(key);
            if (it == m.end()) {
                m.insert(std::map<std::string, Move>::value_type(key, psv[i]));
            } else {
                const Move &oldmove = it->second;
                if (oldmove.kind == SLIDE && oldmove.final_dir != oldmove.dir)
                  it->second = psv[i];
            }
        }
    }
    for (std::map<std::string, Move>::iterator it = m.begin(); it != m.end(); ++it) {
        v.push_back(it->second);
    }
}


static Move get_human_move(const Board &board)
{
    board.display();
  top:
    printf("Your move. > "); fflush(stdout);
    char *line = NULL;
    char *rc = getline_113(&line);
    assert(rc != NULL);
    Move move;
    bool success = move.scan(line, board);
    if (!success) {
        printf("That didn't scan as a move.\n");
    } else if (!is_valid_move(board, move)) {
        printf("That parses, but it's not a legal move.\n");
        success = false;
    }
    free(line);
    if (!success) goto top;
    return move;
}


static int evaluate(const Board &b)
{
    const int attacker = 1-b.attacker;
    if (b.game_over()) {
        return (b.ball_pos.isGoal(attacker) ? +1000 : -1000);
    }
    int advancement = b.ball_on_board ? std::max(b.ball_pos.x, b.ball_pos.y) : 7;
    if (attacker == 1)
      advancement = 14-advancement;
    int dist_from_ball = 0;
    int possession = 0;
    int tacklers = 0;
    int wrong_way_penalty = 0;
    int wall_penalty = 0;
    for (int w=0; w < 2; ++w) {
        Pos effective_ball_pos(w ? 4 : 3, w ? 4 : 3);
        if (b.ball_on_board)
          effective_ball_pos = b.ball_pos;
        for (int s=0; s < 3; ++s) {
            const Penguin &p = b.penguin[w][s];
            Pos effective_p_pos(w ? 8 : -1, w ? 8 : -1);
            if (p.on_board)
              effective_p_pos = p.pos;

            int dx = abs(effective_ball_pos.x - effective_p_pos.x);
            int dy = abs(effective_ball_pos.y - effective_p_pos.y);
            int d = std::max(dx, dy);
            dist_from_ball += (w == attacker ? d : -d);
            if (p.on_board && !p.standing) {
                Pos pd = p.pos + p.dir;
                if (pd.off_board() || pd.isGoal(attacker)) {
                    /* Penalty for facing a wall. */
                    wall_penalty += (w == attacker ? 1 : -1);
                } else if (effective_ball_pos.manhattan_distance(pd) > effective_ball_pos.manhattan_distance(p.pos)) {
                    /* Penalty for facing the wrong way. */
                    wrong_way_penalty += (w == attacker ? 1 : -1);
                }
            }
            if (d == 0)
              possession = (w == attacker ? 1 : -1);

            /* Also give points for penguins that could tackle the ball. */
            if (b.ball_on_board) {
                bool could_get_it = (dx == 0 || dy == 0 || dx == dy);
                could_get_it &= (dx <= 3-s) && (dy <= 3-s);
                /* It's in range, but are we pointed the right way? */
                if (p.standing) {
                    /* Yes! */
                } else if (could_get_it) {
                    Pos ps = effective_p_pos;
                    could_get_it = false;
                    for (int i=0; i < 3-s; ++i) {
                        ps += p.dir;
                        if (ps == effective_ball_pos)
                          could_get_it = true;
                    }
                }
                if (could_get_it)
                  tacklers += (w == attacker ? 1 : -1);
            }
        }
    }
    int score = 4*advancement + 6*possession + 9*tacklers;
    score -= dist_from_ball;
    score -= wrong_way_penalty;
    score -= 4*wall_penalty;
    return score;
}

static void findmoves_randomorder(const Board &b, std::vector<Move> &allmoves)
{
    assert(allmoves.empty());
    get_all_moves(b, allmoves);
    std::random_shuffle(allmoves.begin(), allmoves.end());
}

bool get_alphabeta_move(const Board &b, Move &move)
{
    static AlphaBeta<Board,Move,int> ab(evaluate, (void(*)(Board&, const Move&))apply_move,
                                        findmoves_randomorder, findattacker);
    const int ply = 4;
    int bestvalue;
    ab.depth_first_alpha_beta(b, ply, move, bestvalue, -1000, +1000);
    return (bestvalue > -1000);
}


int main()
{
    printf("--------------------------------------------------\n");
    printf("------------------PENGUIN SOCCER------------------\n");
    printf("--------------------------------------------------\n");
    printf("The board is displayed tilted 45 degrees:  WN\n");
    printf("                                           SE\n");
    printf("The ball is not displayed until a penguin\n");
    printf("reaches one of the four central squares.\n");
    printf("  Enter moves using this syntax:\n");
    printf("1sw = small penguin kicks the ball southwest (or slides\n");
    printf("  southwest, if not holding the ball)\n");
    printf("2s w = papa penguin slides west, then rotates to face west\n");
    printf("3! = big penguin stands up\n");
    printf("  The human player (black) enters from the south and must\n");
    printf("get the ball into the north goal.  The computer player\n");
    printf("(white) enters from the north and aims for the south goal.\n");
    printf("--------------------------------------------------\n");

    srand((unsigned int)time(NULL));
    Board b;
    std::vector<Move> allmoves;
        for (int turns_played = 0; true; ++turns_played) {
            if (b.game_over())
              goto somebody_lost_the_game;
            allmoves.clear();
            get_all_moves(b, allmoves);
            if (allmoves.empty()) {
                printf("Wow, no moves available from this position:\n");
                b.display();
                goto somebody_lost_the_game;
            }
            Move move;
            if (findattacker(b) == 0) {
                move = get_human_move(b);
            } else {
                if (allmoves.size() == 1)
                  move = allmoves[0];
                else if (get_alphabeta_move(b, move))
                  /* move has been selected */ ;
                else
                  move = allmoves[rand() % allmoves.size()];
                printf("Computer (White) moves: %s\n", move.toString(1).c_str());
            }
            assert(is_valid_move(b, move));
            apply_move(b, move);
        }
        /* The game ran overtime; it's probably a draw. Let's just consider
         * a draw as bad as a loss; we want to win! */
        goto game_over_in_general;
        
      somebody_lost_the_game:
        b.display();
        if (b.ball_pos.isGoal(0)) {
            puts("Computer lost the game!");
        } else {
            puts("Human lost the game!");
        }
      game_over_in_general:
    return 0;
}

